/*
** Module   : ChuiWidgetFactory.java
** Abstract :
**
** Copyright (c) 2011-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 SIY 20111221 Created initial version
** 002 CA  20130129 Import change.
*/

package com.goldencode.p2j.ui.chui;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Text;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * CHUI-specific widget factory implementation.
 */
public class ChuiWidgetFactory
implements WidgetFactory
{
   /** Mapping between abstract and UI-specific widget classes. */
   static Map<Class<? extends Widget>, Class<? extends Widget>> classMap = new HashMap<Class<? extends Widget>, Class<? extends Widget>>();
   
   static 
   {
      classMap.put(AlertBox.class, AlertBoxImpl.class);
      classMap.put(Browse.class, BrowseImpl.class);
      classMap.put(BrowseColumn.class, BrowseColumn.class);
      classMap.put(Button.class, ButtonImpl.class);
      classMap.put(ComboBox.class, ComboBoxImpl.class);
      classMap.put(Editor.class, EditorImpl.class);
      classMap.put(FillIn.class, FillInImpl.class);
      classMap.put(Frame.class, FrameImpl.class);
      classMap.put(Label.class, LabelImpl.class);
      classMap.put(RadioButton.class, RadioButtonImpl.class);
      classMap.put(RadioSet.class, RadioSet.class);
      
      classMap.put(ScrollableList.class, ScrollableListImpl.class);
      classMap.put(ScrollPane.class, ScrollPaneImpl.class);
      classMap.put(SelectionListBody.class, SelectionListBodyImpl.class);
      classMap.put(SelectionList.class, SelectionListImpl.class);
      classMap.put(Separator.class, SeparatorImpl.class);
      classMap.put(Skip.class, SkipImpl.class);
      classMap.put(Space.class, Space.class);
      classMap.put(StatusLine.class, StatusLineImpl.class);
      classMap.put(Text.class, TextImpl.class);
      classMap.put(Window.class, WindowImpl.class);      
   }
   
   
   /**
    * Create new {@link Label} instance for given text.
    * 
    * @param   text
    *          Label text. 
    * @return  new {@link Label} instance.
    */
   @Override
   public Label createLabel(String text)
   {
      return  new LabelImpl(text);
   }

   /**
    * Create new button for given container and text.
    * 
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          Button text.
    *          
    * @return  new {@link Button} instance.
    */
   @Override
   public Button createButton(Container container, String text)
   {
      return new ButtonImpl(container, text);      
   }

   /**
    * Instantiate UI-specific version of the specified widget class.
    * 
    * @param   <T>
    *          Widget abstract class type.
    * @param   clazz
    *          Widget abstract class for which UI-specific implementation should
    *          be instantiated.
    * @param   cfg
    *          Widget configuration container.
    *          
    * @return  new UI-specific instance of the requested class. 
    */
   @Override
   public <T extends Widget> T create(Class<T> clazz, ComponentConfig cfg)
   {
      @SuppressWarnings("unchecked")
      Class<T> implClazz = (Class<T>) classMap.get(clazz);
      
      if (implClazz == null) {
         throw new RuntimeException("Attempt to instantiate class" + clazz + 
                                    " which has no proper mapping");
      }
      
      try
      {
         Constructor<T> constructor = implClazz.getConstructor(ComponentConfig.class);
         return constructor.newInstance(cfg);
      }
      catch (Exception e)
      {
         throw new RuntimeException("Attempt to instantiate class" + implClazz + 
         " for class " + clazz + " failed", e);
      }
   }

   /**
    * Create UI-specific {@link ScrollableList} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    * @param   model
    *          Scrollable list model.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollableList}.
    */
   @Override
   public ScrollableList createScrollableList(ComboBox parent,
                                              ListModel<String> model)
   {
      return new ScrollableListImpl(parent, parent.model());
   }

   /**
    * Create UI-specific {@link ScrollPane} implementation.
    * 
    * @param   widget 
    *          Widget which will be scrolled.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollPane}.
    */
   @Override
   public ScrollPane createScrollPane(Widget widget)
   {
      return new ScrollPaneImpl(widget);
   }

   /**
    * Create UI-specific {@link Label} instance as a copy of given {@link Label}.
    * 
    * @param   labelInstance
    *          Source {@link Label} instance.
    *          
    * @return  instance of UI-specific implementation of {@link Label}.
    */
   @Override
   public Label createLabel(Label labelInstance)
   {
      return new LabelImpl(labelInstance);
   }

   /**
    * Create UI-specific {@link RadioButton} instance.
    * @param   config
    *          Parent config. 
    * @param   label
    *          Radio button label. 
    * @param   value
    *          Radio button data type. 
    * 
    * @return  instance of UI-specific implementation of {@link RadioButton}.
    */
   @Override
   public RadioButton createRadioButton(RadioSetConfig config, String label,
                                        BaseDataType value)
   {
      return  new RadioButtonImpl(config,
                                  label, 
                                  value);
   }

   /**
    * Create UI-specific {@link Window} instance.
    * 
    * @return  instance of UI-specific implementation of {@link Window}.
    */
   @Override
   public Window createWindow()
   {
      return new WindowImpl();
   }

   /**
    * Create UI-specific {@link MessageLine} instance.
    * 
    * @return  instance of UI-specific implementation of {@link MessageLine}.
    */
   @Override
   public StatusLine createMessageLine()
   {
      return new MessageLine();
   }

   /**
    * Create UI-specific {@link StatusLine} instance.
    * 
    * @return  instance of UI-specific implementation of {@link StatusLine}.
    */
   @Override
   public StatusLine createStatusLine()
   {
      return new StatusLineImpl();
   }
}
